# export_ifp.py @space_view3d_gta_tools
# 2011 ponz
script_info = "GTA DFF Exporter ( built:2011.7.10 )"

# ***** BEGIN GPL LICENSE BLOCK *****
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or ( at your option ) any later version.
#
# This program is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation, 
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ***** END GPL LICENCE BLOCK *****

#####
# ToDo.
#  - Vertex Colors.
#  - 2nd UV Coords.
#  - Generic Textures.
#  - Material Effect.
#####

import bpy
import os
import time
import struct
from bpy.props import *
from mathutils import *
from math import *
from rna_prop_ui import rna_idprop_ui_prop_get
import ctypes
import re
import sys

## Constants


# RW Version
rw_version = {
	0x0800FFFF : "3.?.?.?( GTA III )",
	0x00000310 : "3.1.0.0( GTA III )",
	0x0C02FFFF : "3.3.0.2( GTA VC )" ,
	0x1003FFFF : "3.4.0.3( GTA VC )" ,
	0x1803FFFF : "3.6.0.3( GTA SA )" }

# Geometry Flags
GEOMETRY_TRISTRIP  = 1
GEOMETRY_POSITIONS = 2
GEOMETRY_UVTEXTURE = 4
GEOMETRY_PRELIT    = 8
GEOMETRY_NORMALS   = 16
GEOMETRY_LIGHT     = 32
GEOMETRY_MODULATE  = 64
GEOMETRY_ETEXTURE  = 128

## Classes
# Data Class
class ClassHeaderEx:
	def __init__( self ):
		self.id   = 0
		self.size = 0
		self.ver  = 0

class ClassCollisions:
	def __init__( self ):
		self.name              = ""
		self.bounding_boxes    = []
		self.bounding_spheres  = []
		self.collision_boxes   = []
		self.collision_spheres = []
		self.collision_meshes  = []
		self.shadow_meshes     = []


## Functions
def SetSection( id, data ):
	section = []
	section.extend( struct.pack('<3I', id, len( data ), 0x1803FFFF ) )
	section.extend( data )
	return section

def GetParentRecursive( obj ):
	if None != obj.parent:
		return GetParentRecursive( obj.parent )
	else:
		return obj

def GetChildrenRecursive( obj ):
	children = [obj]
	for child in obj.children:
		children.extend( GetChildrenRecursive( child ) )
	return children

def GetLinkedObjs( active_obj ):
	root_obj = GetParentRecursive( active_obj )
	return GetChildrenRecursive( root_obj )

def ExportDFF( filepath ):
	dff_props = bpy.context.scene.gta_tools.dff_props
	global script_info
	print( "\n-----\n" + script_info + "\n-----" )
	
	if "CHAR" == dff_props.exp_type:
		dff_props.char_skin = True
		dff_props.vehicle   = False
	elif "VEHICLE" == dff_props.exp_type:
		dff_props.char_skin = False
		dff_props.vehicle   = True
	else:
		dff_props.char_skin = False
		dff_props.vehicle   = False
	mobj = None
	aobj = None
	bones = []
	objs = []
	mobjs = []
	cobjs = ClassCollisions()
	
	active_obj = bpy.context.active_object
	print( "Active Object : " + active_obj.name )
	
	## Check Active Objects for CharSkin
	if dff_props.char_skin:
		if 'ARMATURE' == active_obj.type:
			aobj = active_obj
		elif 'MESH' == active_obj.type:
			aobj = active_obj.parent
		else:
			print( "Error : Active Object is not \"ARMATURE or MESH\"")
			print( "active_obj.name")
			return
		
		## Check Armature Object
		if 'ARMATURE' != aobj.type:
			print( "Error : Selected Mesh Object has no parent Armature")
			print( "active_obj.name")
			return
		print( "Armature Object: " + aobj.name )
		bones = aobj.data.bones
		
		## Check Mesh Object
		c_mobjs = []
		for obj in aobj.children:
			if 'MESH' == obj.type:
				c_mobjs.append( obj )
				print( "Mesh Object : " + obj.name )
		if 0 == len( c_mobjs ):
			print( "Mesh Object : None" )
		if 1 != len( c_mobjs ):
			print( "Error : Armature Object must have one Child Mesh Object.")
			return
		mobj = c_mobjs[0]
		
		## Check Armature Modifire
		amods = []
		for mod in mobj.modifiers:
			if( 'ARMATURE' == mod.type ):
				amods.append( mod )
				print( "Armature Modifire : " + mod.name )
		if 0 == len( amods ):
			print( "Armature Modifire : None" )
		if 1 != len( amods ):
			print( "Error : Mesh Object must have one Armature Modifire.")
			return
		objs.append( mobj )
		mobjs.append(mobj)
	
	## Check Active Objects for Non-CharSkin
	else:
		tmp_objs = GetLinkedObjs( active_obj )
		
		for obj in tmp_objs:
			obj_name = obj.name.split( "." )[0]
			## Collisions for Vehicle Models
			if "bounding_box" == obj_name:
				if 'EMPTY'  == obj.type:
					if 'CUBE' == obj.empty_draw_type:
						cobjs.bounding_boxes.append( obj )
			elif "bounding_sphere" == obj_name:
				if 'EMPTY'  == obj.type:
					if 'SPHERE' == obj.empty_draw_type:
						cobjs.bounding_spheres.append( obj )
			elif "collision_box" == obj_name:
				if 'EMPTY'  == obj.type:
					if 'CUBE' == obj.empty_draw_type:
						cobjs.collision_boxes.append( obj )
			elif "collision_sphere" == obj_name:
				if 'EMPTY'  == obj.type:
					if 'SPHERE' == obj.empty_draw_type:
						cobjs.collision_spheres.append( obj )
			elif "collision_mesh" == obj_name:
				if 'MESH'  == obj.type:
					cobjs.collision_meshes.append( obj )
			elif "shadow_mesh" == obj_name:
				if 'MESH'  == obj.type:
					cobjs.shadow_meshes.append( obj )
			
			## Mesh, Nodes
			else:
				if 'MESH' == obj.type:
					print( "Mesh Object : " + obj.name )
					objs.append(obj)
					mobjs.append(obj)
				elif 'EMPTY' == obj.type:
					print( "Node Object : " + obj.name )
					objs.append(obj)
		
		if dff_props.vehicle:
			## Check Collisions
			if 0 == len( cobjs.bounding_boxes ):
				bpy.context.scene.gta_tools.msg += ( "Error : No \"bounding_box\" for vehicle collision data." )
				return
			if 0 == len( cobjs.bounding_spheres ):
				bpy.context.scene.gta_tools.msg += ( "Error : No \"bounding_sphare\" for vehicle collision data." )
				return
			if 0 == len( cobjs.collision_boxes ) + len( cobjs.collision_spheres ) + len( cobjs.collision_meshes ):
				bpy.context.scene.gta_tools.msg += ( "Error : No Collisions for vehicle collision data." )
				return
	
	## Set DFF Data
	dff = []
	clump = []
	structure = struct.pack('<3i', len( mobjs ), 0, 0 )
	clump.extend( SetSection( 0x01, structure ) )
	
	# Frame list @Clump
	framlist = []
	if dff_props.char_skin:
		num_frams = len( bones ) + 1
	else:
		num_frams = len( objs )
	
	# Struct @Frame list
	structure = []
	structure.extend( struct.pack('<I', num_frams ) )
	
	# for Mesh, Nodes
	for obj in objs:
		pfram = -1
		if None != obj.parent:
			if obj.parent in objs:
				pfram = objs.index(obj.parent)
		if -1 == pfram:
			mcf = 0x00020003
			if dff_props.vehicle:
				cobjs.name = obj.name.split( "." )[0]
		else:
			mcf = 0x00000003
		structure.extend( struct.pack('<3f', *obj.rotation_quaternion.to_matrix()[0] ) )
		structure.extend( struct.pack('<3f', *obj.rotation_quaternion.to_matrix()[1] ) )
		structure.extend( struct.pack('<3f', *obj.rotation_quaternion.to_matrix()[2] ) )
		structure.extend( struct.pack('<3f', *obj.location ) )
		structure.extend( struct.pack('<2i', pfram, mcf ) )
	
	# for Bones
	if dff_props.char_skin:
		for bone in bones:
			pbone = bone.parent
			if None == pbone:
				pfram = 0
				tail_ofs = Vector( ( 0.0, 0.0, 0.0 ) )
			else:
				pfram = list(bones).index( pbone ) + 1
				tail_ofs = Vector( ( 0.0, 0.05, 0.0 ) )
			
			mcf = 0x00000003
			structure.extend( struct.pack('<3f', *bone.matrix[0] ) )
			structure.extend( struct.pack('<3f', *bone.matrix[1] ) )
			structure.extend( struct.pack('<3f', *bone.matrix[2] ) )
			structure.extend( struct.pack('<3f', *bone.head + tail_ofs ) )
			structure.extend( struct.pack('<2i', pfram, mcf ) )
	
	framlist.extend( SetSection( 0x01, structure ) )
	
	# Extention @Frame list
	if dff_props.char_skin:
		framlist.extend( SetSection( 0x03, [] ) )
		bone_dict = {}
		for ibone, bone in enumerate( bones ):
			ext = []
			# HAnim PLG @Extention
			haplg = []
			if 0 == ibone:
				nbone = len( bones )
			else:
				nbone = 0
			haplg.extend( struct.pack('<3I', 256, bone["bone_id"], nbone ) )
			bone_dict[bone.name] = [ibone,0]
			
			if 0 != nbone:
				haplg.extend( struct.pack('<2I', 0, 36 ) )
				for ibone_0, bone_0 in enumerate( bones ):
					btype = bone_0["bone_type"]
					haplg.extend( struct.pack('<3I', bone_0["bone_id"], ibone_0, btype ) )
			
			ext.extend( SetSection( 0x11e, haplg ) )
			
			# Frame @Extention
			fram = []
			if dff_props.rev_bone:
				bone_name = bone["org_name"]
			else:
				bone_name = bone.name
			fram.extend( struct.pack('<%ds' %( len( bone_name ) ), bytes( bone_name.encode() ) ) )
			ext.extend( SetSection( 0x253f2fe, fram ) )
			
			framlist.extend( SetSection( 0x03, ext ) )
	else:
		for obj in objs:
			ext = []
			# Frame @Extention
			fram = []
			fram.extend( struct.pack('<%ds' %( len( obj.name ) ), bytes( obj.name.encode() ) ) )
			ext.extend( SetSection( 0x253f2fe, fram ) )
			
			framlist.extend( SetSection( 0x03, ext ) )
	
	clump.extend( SetSection( 0x0E, framlist ) )
	
	# Geometry List @Clump
	geomlist = []
	
	# Struct @Geometry List
	structure = []
	num_geoms = len( mobjs )
	structure.extend( struct.pack('<I', num_geoms ) )
	geomlist.extend( SetSection( 0x01, structure ) )
	
	# Geometry @Geometry List
	for mobj in mobjs:
		print( "-----\nGeometry : " + str( mobj.name ) )
		m = mobj.data
		geom = []
		
		# Struct @Geometry
		structure = []
		nuv = 1
		unk = 0
		num_faces = len( m.faces )
		num_verts = len( m.vertices )
		num_frams = 1
		
		faces = []
		verts = []
		norms = []
		uvs = [0.0]*num_verts*2
		uvdict = {}
		splitdict = {}
		bvis = []
		vws = []
		
		## Geometry Flags
		# for now, Flags are fixed.
		flg_tristrip  = False
		flg_positions = True
		flg_uvtexture = True
		flg_prelit    = False
		flg_normals   = True
		flg_light     = True
		flg_modulate  = not dff_props.char_skin
		flg_etexture  = False
		geom_flags = 0
		if flg_tristrip  : geom_flags |= GEOMETRY_TRISTRIP
		if flg_positions : geom_flags |= GEOMETRY_POSITIONS
		if flg_uvtexture : geom_flags |= GEOMETRY_UVTEXTURE
		if flg_prelit    : geom_flags |= GEOMETRY_PRELIT
		if flg_normals   : geom_flags |= GEOMETRY_NORMALS
		if flg_light     : geom_flags |= GEOMETRY_LIGHT
		if flg_modulate  : geom_flags |= GEOMETRY_MODULATE
		if flg_etexture  : geom_flags |= GEOMETRY_ETEXTURE
		
		print( " Geometry Flags   : " + hex( geom_flags ) )
		print( " Blender Faces    : " + str( num_faces ) + " faces" )
		print( " Blender Vertices : " + str( num_verts ) + " verts" )
		
		## set vert params
		for i, v in enumerate( m.vertices ):
			verts += list( v.co )
			norms += list( v.normal )
		
		if dff_props.char_skin:
			# set weight params
			for i, v in enumerate( m.vertices ):
				bvi = [0, 0, 0, 0]
				vw = [0.0, 0.0, 0.0, 0.0]
				vwtot = 0
				for ig, bvg in enumerate( v.groups ):
					vgname = mobj.vertex_groups[bvg.group].name
					if 4 == ig:
						break
					if vgname in bone_dict:
						bvi[ig] = bone_dict[vgname][0]
						vw[ig] = bvg.weight
						vwtot += bvg.weight
						bone_dict[vgname][1] += 1
				bvis += bvi
				for w in vw:
					if 0.001 < w: w /= vwtot
					else: w = 0.0
				vws  += vw
		
		## set face params
		# fix quad face to triangle face in this block.
		# add verts for uv face in this block.
		uvf = m.uv_textures[0].data
		for iface, f in enumerate( m.faces ):
			fvi = list( f.vertices )
			for i in range( len( f.vertices ) ):
				u = uvf[iface].uv[i][0]
				v = uvf[iface].uv[i][1]
				if f.vertices[i] in uvdict:
					ref_uv = uvdict[f.vertices[i]]
					matched = False
					while( True ):
						delta = Vector( ( u - ref_uv[0], v - ref_uv[1] ) ).length
						if( ( dff_props.uv_th >= delta ) and ( f.material_index == ref_uv[2] ) ):
							matched = True
							fvi[i] = ref_uv[3]
							break
						elif( 0 == len( ref_uv[4] ) ):
							break
						else:
							ref_uv = ref_uv[4]
					if False == matched:
						uvs.extend( ( u, 1 - v ) )
						#uvs.extend( ( -1,-1 ) )
						verts.extend( verts[f.vertices[i]*3:f.vertices[i]*3+3] )
						norms.extend( norms[f.vertices[i]*3:f.vertices[i]*3+3] )
						if dff_props.char_skin:
							bvis.extend ( bvis [f.vertices[i]*4:f.vertices[i]*4+4] )
							vws.extend  ( vws  [f.vertices[i]*4:f.vertices[i]*4+4] )
						fvi[i] = num_verts
						num_verts += 1
						ref_uv[4] = [ u, v, f.material_index, fvi[i], [] ]
				else:
					uvdict[f.vertices[i]] = [ u, v, f.material_index, f.vertices[i], [] ]
					uvs[f.vertices[i]*2]   = u
					uvs[f.vertices[i]*2+1] = 1 - v
			faces.extend( ( fvi[1], fvi[0], f.material_index, fvi[2] ) )
			
			if not( f.material_index in splitdict ):
				splitdict[f.material_index] = []
			splitdict[f.material_index].extend( struct.pack("<3i", fvi[0], fvi[1], fvi[2] ) )
			
			if 4 == len( fvi ):
				num_faces += 1
				faces.extend( ( fvi[2], fvi[0], f.material_index, fvi[3] ) )
				
				if not( f.material_index in splitdict ):
					splitdict[f.material_index] = []
				splitdict[f.material_index].extend( struct.pack("<3i", fvi[0], fvi[2], fvi[3] ) )
		
		print( " DFF Faces    : " + str( num_faces ) + " faces" )
		print( " DFF Vertices : " + str( num_verts ) + " verts" )
		
		## set boundary params
		bounds = [0.0]*6
		for v in m.vertices:
			bounds[0] = min( bounds[0], v.co[0] )
			bounds[1] = max( bounds[1], v.co[0] )
			bounds[2] = min( bounds[2], v.co[1] )
			bounds[3] = max( bounds[3], v.co[1] )
			bounds[4] = min( bounds[4], v.co[2] )
			bounds[5] = max( bounds[5], v.co[2] )
		bcent = [( bounds[1] + bounds[0] ) / 2, ( bounds[3] + bounds[2] ) / 2, ( bounds[5] + bounds[4] ) / 2]
		brad = sqrt( ( bounds[1] - bounds[0] ) ** 2 + ( bounds[3] - bounds[2] ) ** 2 + ( bounds[5] - bounds[4] ) ** 2 ) / 2
		
		## set Geometry Struct
		structure.extend( struct.pack('<H2B3I', geom_flags, nuv, unk, num_faces, num_verts, num_frams ) )
		structure.extend( struct.pack( "<%df" %( num_verts*2 ), *uvs ) )                      # uv coods
		structure.extend( struct.pack( "<%dH" %( num_faces*4 ), *faces ) )                    # vert ids of faces
		structure.extend( struct.pack( "<4f2i", bcent[0], bcent[1], bcent[2], brad, 1, 1 ) )  # boundary
		structure.extend( struct.pack( "<%df" %( num_verts*3 ), *verts ) )                    # vert coods
		structure.extend( struct.pack( "<%df" %( num_verts*3 ), *norms ) )                    # normal vectors
		
		geom.extend( SetSection( 0x01, structure ) )
		
		## Material List @Geometry
		matlist = []
		
		## Struct @Material List
		structure = []
		nmat = len( m.materials )
		#print("nmat:",nmat)
		reserved_array = [-1]*nmat
		structure.extend( struct.pack( "<I%di" %( nmat ), nmat, *reserved_array ) )
		matlist.extend( SetSection( 0x01, structure ) )
		
		## Material @Material List
		for mat in m.materials:
			material = []
			
			## Struct @Material
			structure = []
			col = mat.diffuse_color
			alpha = mat.alpha
			ntex = 0
			for tslot in mat.texture_slots:
				if None != tslot: ntex += 1
			unks = [1.0]*3
			structure.extend( struct.pack( "<I4B2I3f", 0, int( col[0]*255 ), int( col[1]*255 ), int( col[2]*255 ), int( alpha*255 ), 1, ntex, *unks ) )
			material.extend( SetSection( 0x01, structure ) )
			
			## Texture @Material
			for tslot in mat.texture_slots:
				if None != tslot:
					texture = []
					
					## Struct @Texture
					structure = []
					structure.extend( struct.pack( "<2H", 0x106, 1 ) )
					texture.extend( SetSection( 0x01, structure ) )
					
					## String ( texture ) @Texture
					texname = []
					if None != tslot.texture.image:
						tmpstr = tslot.texture.image.name.split("\\")[-1].split(".")[0]
					elif "color_map" in tslot.texture.keys():
						tmpstr = tslot.texture["color_map"]
					else:
						tmpstr = tslot.texture.name.split(".")[0]
					binstr = "%dsB0i"%(len(tmpstr),)
					texname.extend( struct.pack( binstr, bytes( tmpstr.encode() ), 0 ) )
					texture.extend( SetSection( 0x02, texname ) )
					
					## String ( mask ) @Texture
					texname = []
					tmpstr = ""
					binstr = "%dsB0i"%(len(tmpstr),)
					texname.extend( struct.pack( binstr, bytes( tmpstr.encode() ), 0 ) )
					texture.extend( SetSection( 0x02, texname ) )
					
					## Extention @Texture
					ext = []
					texture.extend( SetSection( 0x03, ext ) )
					material.extend( SetSection( 0x06, texture ) )
			
			## Extention @Material
			ext = []
			
			if not dff_props.char_skin:
				# Reflection Materil
				refmat = []
				refmat.extend( struct.pack( "<3f", *mat.specular_color ) )
				refmat.extend( struct.pack( "<2fI", mat.specular_alpha, mat.specular_intensity, 0 ) )
				ext.extend( SetSection( 0x253f2fc, refmat ) )
				
				# Specular Materil
				if "GTAMAT_spec_level" in mat.keys():
					specmat = []
					specmat.extend( struct.pack( "<f", mat["GTAMAT_spec_level"] ) )
					spec_tex = ""
					if "GTAMAT_spec_texture" in mat.keys():
						spec_tex = mat["GTAMAT_spec_texture"]
					specmat.extend( struct.pack( "<%ds2I" %len( spec_tex ), bytes( spec_tex.encode() ), 0, 0 ) )
					ext.extend( SetSection( 0x253f2f6, specmat ) )
			
			material.extend( SetSection( 0x03, ext ) )
			matlist.extend( SetSection( 0x07, material ) )
		
		geom.extend( SetSection( 0x08, matlist ) )
		
		## Extention @Materila List
		ext = []
		
		## Bin Mesh PLG @Extention
		bmplg = []
		matsplit = []
		fc = 0
		for matid in splitdict:
			splitlist = splitdict[matid]
			matsplit.extend( struct.pack("<2I", len(splitlist) // 4, matid) )
			matsplit.extend( splitlist )
			fc += len(splitlist) // 4
		bmplg.extend( struct.pack("<3I", 0, len(splitdict), fc ) )
		bmplg.extend( matsplit )
		ext.extend( SetSection( 0x50e, bmplg ) )
		
		if dff_props.char_skin:
			## Skin PLG @Extention
			skinplg = []
			spbones = []
			ibmats = []
			spunk = 4
			
			for vg in mobj.vertex_groups:
				if vg.name in bone_dict:
					if 0 < bone_dict[vg.name][1]:
						spbones.append( bone_dict[vg.name][0] )
			
			for ibone, bone in enumerate( bones ):
				ibmat = bone.matrix_local.inverted()
				ibmat[3][3] = 0.0  # ???
				ibmats.extend( ibmat[0] )
				ibmats.extend( ibmat[1] )
				ibmats.extend( ibmat[2] )
				ibmats.extend( ibmat[3] )
			
			skinplg.extend( struct.pack("<4B", len( bones ), len( spbones ), spunk, 0 ) )
			skinplg.extend( struct.pack("<%dB" %len( spbones ), *spbones ) )
			skinplg.extend( struct.pack("<%dB" %len( bvis ), *bvis ) )
			skinplg.extend( struct.pack("<%df" %len( vws ), *vws ) )
			skinplg.extend( struct.pack("<%df" %len( ibmats ), *ibmats ) )
			skinplg.extend( struct.pack("<3i", 0,0,0 ) )  # ?????
			
			ext.extend( SetSection( 0x116, skinplg ) )
		
		## Mesh Extension @Extention
		meshext = []
		meshext.extend( struct.pack("<i", 0 ) )
		ext.extend( SetSection( 0x253f2fd, meshext ) )
		geom.extend( SetSection( 0x03, ext ) )
		
		geomlist.extend( SetSection( 0x0F, geom ) )
	
	clump.extend( SetSection( 0x1A, geomlist ) )
	
	## Atomic @Clump
	for mobj in mobjs:
		atom = []
		
		## Struct @Atmic
		structure = []
		ifram = objs.index(mobj)
		igeom = mobjs.index(mobj)
		structure.extend( struct.pack( "<4I", ifram, igeom, 5, 0 ) )
		atom.extend( SetSection( 0x01, structure ) )
		
		## Extension @Atmic
		ext = []
		
		## Right To Render @Atmic
		rtr = []
		rtr.extend( struct.pack( "<2I", 0x116, 1 ) )
		ext.extend( SetSection( 0x1f, rtr ) )
		
		atom.extend( SetSection( 0x03, ext ) )
		
		clump.extend( SetSection( 0x14, atom ) )
		
	
	## Extension @Clump
	ext = []
	
	### Collision Model @Extension
	if dff_props.vehicle:
		print( "-----\nCollisions" )
		## TBounds @Collision Model
		tbounds = []
		if 0 == len( cobjs.bounding_boxes ):
			pass
		else:
			bbox = cobjs.bounding_boxes[0]
			bbcent = bbox.location
			bbscale = bbox.scale * bbox.empty_draw_size
			bbmin = bbcent - bbscale
			bbmax = bbcent + bbscale
		
		if 0 == len( cobjs.bounding_spheres ):
			pass
		else:
			bsphere = cobjs.bounding_spheres[0]
			bscent = bsphere.location
			bsscale = bsphere.scale * bsphere.empty_draw_size
			bsrad = max( bsscale )
		
		tbounds.extend( struct.pack( "<3f", *bbmin ) )
		tbounds.extend( struct.pack( "<3f", *bbmax ) )
		tbounds.extend( struct.pack( "<3f", *bscent ) )
		tbounds.extend( struct.pack( "<f", bsrad ) )
		
		print( " TBounds :" )
		print( "  ", bbmin )
		print( "  ", bbmax )
		print( "  ", bscent )
		print( "  ", brad )
		
		## TSphere @Collision Model
		cspheres = []
		for csphere in cobjs.collision_spheres:
			cscent = csphere.location
			csscale = csphere.scale * csphere.empty_draw_size
			csrad = max( csscale )
			cssurf = [ 0, 0, 0, 0 ]
			if "material"   in csphere.keys(): cssurf[0] = csphere["material"]
			if "flag"       in csphere.keys(): cssurf[1] = csphere["flag"]
			if "brightness" in csphere.keys(): cssurf[2] = csphere["brightness"]
			if "light"      in csphere.keys(): cssurf[3] = csphere["light"]
			cspheres.extend( struct.pack( "<3f", *cscent ) )
			cspheres.extend( struct.pack( "<f" , csrad ) )
			cspheres.extend( struct.pack( "<4B", *cssurf ) )
		
		print( " TSpheres :", len( cobjs.collision_spheres ), len( cspheres ) )
		
		## TBox @Collision Model
		cboxes = []
		for cbox in cobjs.collision_boxes:
			cbcent = cbox.location
			cbscale = cbox.scale * cbox.empty_draw_size
			cbmin = cbcent - cbscale
			cbmax = cbcent + cbscale
			cbsurf = [ 0, 0, 0, 0 ]
			if "material"   in cbox.keys(): cbsurf[0] = cbox["material"]
			if "flag"       in cbox.keys(): cbsurf[1] = cbox["flag"]
			if "brightness" in cbox.keys(): cbsurf[2] = cbox["brightness"]
			if "light"      in cbox.keys(): cbsurf[3] = cbox["light"]
			cboxes.extend( struct.pack( "<3f", *cbmin ) )
			cboxes.extend( struct.pack( "<3f", *cbmax ) )
			cboxes.extend( struct.pack( "<4B", *cbsurf ) )
		
		print( " TBoxes :", len( cobjs.collision_boxes ), len( cboxes ) )
		
		### Collision Mesh
		cverts = []
		cfaces = []
		## TVertex, TFace @Collision Model
		if 0 < len( cobjs.collision_meshes ):
			mobj = cobjs.collision_meshes[0]
			m = mobj.data
			# TVertex @Collision Mesh
			for vert in m.vertices:
				cverts.extend( struct.pack( "<3h", int( vert.co[0]* 128.0 ), int( vert.co[1]* 128.0 ), int( vert.co[2]* 128.0 ) ) )
			
			# TFace @Collision Mesh
			mats = []
			for material in m.materials:
				mat = [0, 255]
				if "GTAMAT_coll_material" in material.keys(): mat[0] = material["GTAMAT_coll_material"]
				if "GTAMAT_coll_light"    in material.keys(): mat[1] = material["GTAMAT_coll_light"]
				mats.append( mat )
			for face in m.faces:
				cfaces.extend( struct.pack( "<3H", face.vertices[0], face.vertices[1], face.vertices[2] ) )
				cfaces.extend( struct.pack( "<2B", *mats[face.material_index] ) )
				if 4 == len( face.vertices ):
					cfaces.extend( struct.pack( "<3H", face.vertices[0], face.vertices[2], face.vertices[3] ) )
					cfaces.extend( struct.pack( "<2B", *mats[face.material_index] ) )
		
		print( " Collision Meshes :", len( cobjs.collision_meshes ), len( cverts ), len( cfaces ) )
		
		### Shadow Mesh
		sverts = []
		sfaces = []
		## TVertex, TFace @Collision Model
		if 0 < len( cobjs.shadow_meshes ):
			mobj = cobjs.shadow_meshes[0]
			m = mobj.data
			# TVertex @Shadow Mesh
			for vert in m.vertices:
				sverts.extend( struct.pack( "<3h", int( vert.co[0]* 128.0 ), int( vert.co[1]* 128.0 ), int( vert.co[2]* 128.0 ) ) )
			
			# TFace @Shadow Mesh
			mats = []
			for material in m.materials:
				mat = [0, 255]
				if "GTAMAT_coll_material" in material.keys(): mat[0] = material["GTAMAT_coll_material"]
				if "GTAMAT_coll_light"    in material.keys(): mat[1] = material["GTAMAT_coll_light"]
				mats.append( mat )
			for face in m.faces:
				sfaces.extend( struct.pack( "<3H", face.vertices[0], face.vertices[1], face.vertices[2] ) )
				sfaces.extend( struct.pack( "<2B", *mats[face.material_index] ) )
				if 4 == len( face.vertices ):
					sfaces.extend( struct.pack( "<3H", face.vertices[0], face.vertices[2], face.vertices[3] ) )
					sfaces.extend( struct.pack( "<2B", *mats[face.material_index] ) )
		
		print( " Shadow Meshes :", len( cobjs.shadow_meshes ), len( sverts ), len( sfaces ) )
		
		## Header @Collision Model
		coll = []
		fourcc = 0x334c4f43  # "COL3"
		size = len( cspheres ) + len( cboxes ) + len( cverts ) + len( cfaces ) + len( sverts ) + len( sfaces ) + 112  # Header Size of COL3(120) - 8
		coll_name = cobjs.name
		print( len( cspheres ) , len( cboxes ) , len( cverts ) , len( cfaces ) , len( sverts ) , len( sfaces ) )  # Header Size of COL3(120) - 8
		
		flags = 0
		if 0 < len( cobjs.collision_boxes ) + len( cobjs.collision_spheres ) + len( cobjs.collision_meshes ):
			flags |= 0x02
		if 0 < len( cobjs.shadow_meshes ):
			flags |= 0x10
		
		ofs_csphere = 116  # header size(120) - 4
		ofs_cbox    = ofs_csphere + len( cspheres )
		ofs_cverts  = ofs_cbox    + len( cboxes   )
		ofs_cfaces  = ofs_cverts  + len( cverts   )
		ofs_sverts  = ofs_cfaces  + len( cfaces   )
		ofs_sfaces  = ofs_sverts  + len( sverts   )
		
		coll.extend( struct.pack( "<2I20sI", fourcc, size, bytes( coll_name.encode() ), 0 ) )
		coll.extend( tbounds )
		coll.extend( struct.pack( "<2H2I", len( cobjs.collision_spheres ), len( cobjs.collision_boxes ), len( cfaces ) // 8, flags ) )
		coll.extend( struct.pack( "<3I", ofs_csphere, ofs_cbox, 0 ) )
		coll.extend( struct.pack( "<3I", ofs_cverts, ofs_cfaces, 0 ) )
		coll.extend( struct.pack( "<3I", len( sfaces ) // 8, ofs_sverts, ofs_sfaces ) )
		
		## Body @Collision Model
		coll.extend( cspheres )
		coll.extend( cboxes   )
		coll.extend( cverts   )
		coll.extend( cfaces   )
		coll.extend( sverts   )
		coll.extend( sfaces   )
		
		ext.extend( SetSection( 0x253f2fa, coll ) )
	
	clump.extend( SetSection( 0x03, ext ) )
	
	dff.extend( SetSection( 0x10, clump ) )
	
	## Open File
	try:
		file = open( filepath, "wb" )
	except:
		pass#print( "-----\nError : failed to open " + filepath )
		return
	
	file.write( struct.pack('<%dB' %( len( dff ) ), *dff ) )
	file.close()
	
	bpy.context.scene.gta_tools.msg +=  "Finished : Successfully.\n"
	print( "Finished : Successfully." )
	return
