# import_dff.py @space_view3d_gta_tools
# 2011 ponz
script_info = "GTA DFF Importer ( built:2011.02.05 )"

# ***** BEGIN GPL LICENSE BLOCK *****
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or ( at your option ) any later version.
#
# This program is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation, 
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ***** END GPL LICENCE BLOCK *****

#####
# ToDo.
#  - Vertex Colors.
#  - 2nd UV Coords.
#  - Generic Textures.
#  - Material Effect.
#####

import bpy
import os
import time
import struct
if bpy.app.version[1] < 58:
	from io_utils import load_image
else:
	from bpy_extras.image_utils import load_image
from bpy.props import *
from mathutils import *
from math import *
from rna_prop_ui import rna_idprop_ui_prop_get
import ctypes
import re
import sys

## Constants
# RW Version
rw_version = {
	0x0800FFFF : "3.?.?.?( GTA III )",
	0x00000310 : "3.1.0.0( GTA III )",
	0x0C02FFFF : "3.3.0.2( GTA VC )" ,
	0x1003FFFF : "3.4.0.3( GTA VC )" ,
	0x1803FFFF : "3.6.0.3( GTA SA )" }

# Geometry Flags
GEOMETRY_TRISTRIP  = 1
GEOMETRY_POSITIONS = 2
GEOMETRY_UVTEXTURE = 4
GEOMETRY_PRELIT    = 8
GEOMETRY_NORMALS   = 16
GEOMETRY_LIGHT     = 32
GEOMETRY_MODULATE  = 64
GEOMETRY_ETEXTURE  = 128

# Fixed Bone Name
fixed_bone_name = {
	   0 : "Root",
	   1 : "Pelvis",
	   2 : "Spine 1",
	   3 : "Spine 2",
	   4 : "Neck",
	   5 : "Head",
	   6 : "Brow L",
	   7 : "Brow R",
	   8 : "Jaw",
	  21 : "Bip01 Clavicle R",
	  22 : "UpperArm R",
	  23 : "ForeArm R",
	  24 : "Hand R",
	  25 : "Finger R",
	  26 : "Finger01 R",
	  28 : "Thumb1 R",
	  29 : "Thumb2 R",
	  30 : "Lip11 L",
	  31 : "Bip01 Clavicle L",
	  32 : "UpperArm L",
	  33 : "ForeArm L",
	  34 : "Hand L",
	  35 : "Finger L",
	  36 : "Finger01 L",
	  38 : "Thumb1 L",
	  39 : "Thumb2 L",
	  40 : "Jaw22",
	  41 : "Thigh L",
	  42 : "Calf L",
	  43 : "Foot L",
	  44 : "Toe0 L",
	  51 : "Thigh R",
	  52 : "Calf R",
	  53 : "Foot R",
	  54 : "Toe0 R",
	 201 : "Belly",
	 301 : "Breast R",
	 302 : "Breast L",
	5001 : "Brow1 R",
	5002 : "Brow2 R",
	5003 : "Brow2 L",
	5004 : "Brow1 L",
	5005 : "Lid R",
	5006 : "Lid L",
	5007 : "Tlip3 R",
	5008 : "Tlip3 L",
	5009 : "Tlip1 R",
	5010 : "Tlip2 R",
	5011 : "Tlip1 L",
	5012 : "Tlip2 L",
	5013 : "Corner R",
	5014 : "Corner L",
	5015 : "Jaw1",
	5016 : "Jaw2",
	5017 : "Lip1 L",
	5018 : "Eye R",
	5019 : "Eye L",
	5020 : "Cheek R",
	5021 : "Cheek L",
	5022 : "HeadNub",
	5023 : "Finger0Nub L",
	5024 : "Finger0Nub R",
	5025 : "Toe0Nub L",
	5026 : "Toe0Nub R"}


## Classes
# Data Class
class ClassHeader:
	def __init__( self, header, pos ):
		self.id   = header[0]
		self.size = header[1]
		self.ver  = header[2]
		self.ofs  = pos

class ClassTexture:
	def __init__( self ):
		self.flags     = 0
		self.name_tex  = ""
		self.name_mask = ""

class ClassMaterial:
	def __init__( self ):
		self.color = []
		self.texs  = []
		self.ref   = []  # [ R, G, B, A, Intensity ]
		self.spec  = []  # [ Level, Texture Name ]

class ClassMSplit:
	def __init__( self ):
		self.matid = 0
		self.vids  = []

class ClassFrame:
	def __init__( self ):
		self.rot     = []
		self.pos     = []
		self.pfram   = []
		self.integer = []
		self.name    = ""
		self.bid     = -1
		self.btable  = []  #  Bone Info in HAnimPLG of Root Bone: [ Bone ID, Bone No, Type ]

class ClassGeometry:
	def __init__( self ):
		self.flags  = 0
		self.verts  = []
		self.vcols  = []
		self.uvs    = []
		self.norms  = []
		self.faces  = []
		self.mids   = []
		self.fmats  = []
		self.mats   = []
		self.ftype  = 0
		self.mspls  = []
		self.spunk  = 0
		self.spids  = []
		self.bvis   = []
		self.vws    = []
		self.ibmats = []
		self.uvfaces = []

class ClassAtomic:
	def __init__( self ):
		self.ifram   = 0
		self.igeom   = 0

class ClassCollision:
	def __init__( self ):
		self.fourcc   = 0
		self.name     = ""
		self.size     = 0
		self.bbox     = [Vector( ( 0.0, 0.0, 0.0 ) ), Vector( ( 0.0, 0.0, 0.0 ) )]
		self.bsphere  = [Vector( ( 0.0, 0.0, 0.0 ) ), 0]
		self.boxes    = []
		self.spheres  = []
		self.verts    = []
		self.faces    = []
		self.surfs    = []
		self.sverts   = []
		self.sfaces   = []
		self.ssurfs   = []

class ClassClump:
	def __init__( self ):
		self.name  = ""
		self.frams = []
		self.geoms = []
		self.atoms = []
		self.coll  = ClassCollision()

class ClassDFF:
	def __init__( self ):
		self.ini      = 0
		self.end      = 0
		self.ver_id   = 0
		self.clumps   = []
		self.name     = ""
		self.pos      = Vector( ( 0.0, 0.0, 0.0 ) )
		self.rot      = Quaternion( ( 1.0, 0.0, 0.0, 0.0 ) )
		self.skip_nodes = False
		self.tex_dict = {}

## Functions
def ReadHeader( file ):
	return ClassHeader( struct.unpack( "<iii", file.read( 12 ) ), file.tell() )

def ReadHeaders( file, endof_section, id_sec = -1 ):
	cur_pos = file.tell()
	headers = []
	while endof_section >= file.tell() + 12:
		h = ReadHeader( file )
		if h.ver in rw_version:
			if endof_section >= h.ofs + h.size:
				if -1 == id_sec or id_sec == h.id:
					headers.append( h )
				file.seek( h.size, 1 )
				continue
		break
	file.seek( cur_pos, 0 )
	return headers

def CheckHeader( file, id_sec ):
	h = ReadHeader( file )
	if h.id != id_sec:
		bpy.context.scene.gta_tools.msg +=   "Wrong Section ID @" + hex( h.ofs - 12 ) + " : expected " + hex( id_sec ) + "\n"
		raise ImportError
	return h

def CheckMultiByte( str ):
	for c in str:
		if ord( c ) > 255: return True
	return False

def VecRotate( vec, quat ):  # ( Vector, Quaternion )
	vq = Quaternion( [0.0, vec[0], vec[1], vec[2]] )
	vq = quat * vq * quat.inverted()
	return Vector( vq[1:] )

def VecRotateMat( vec, quat ):  # ( Vector, Quaternion )
	v = vec.copy()
	v.rotate(quat)
	return v

def AlignRoll( vec, vecz, tarz ):  # ( Vector, Vector, Vector )
	sine_roll = vec.normalized().dot(vecz.normalized().cross( tarz.normalized() ) )
	if 1 < abs(sine_roll) : sine_roll /= abs(sine_roll)
	if 0 < vecz.dot( tarz ): return asin( sine_roll )
	elif 0 < sine_roll             : return -asin( sine_roll ) + pi
	else                           : return -asin( sine_roll ) - pi

def GetWRot( rot, refrot ):
	return rot.copy.rotate( refrot)

def GetWPos( pos, refpos ):
	return pos + refpos

def SetLayer( obj, ilayers ):
	obj_layers = [False]*20
	for ilayer in ilayers:
		obj_layers[ilayer] = True
	obj.layers = obj_layers

def MaterialSplit( geom ):
	geom.faces = []
	geom.mids  = []
	
	if 1 == geom.ftype:  # ftype = 1: Triangle Strip
		for mspl in geom.mspls:
			par = 1  # par: parity to decide Normal-Vector direction, set first face to Right-Handed
			for fi in range( len( mspl.vids ) - 2 ):
				f = mspl.vids[fi:fi+3]
				if f[0] != f[1] and f[1] != f[2] and f[2] != f[0]:
					if 1 == par: geom.faces += [ mspl.vids[fi], mspl.vids[fi+1], mspl.vids[fi+2], 0 ]
					else       : geom.faces += [ mspl.vids[fi+1], mspl.vids[fi], mspl.vids[fi+2], 0 ]
					geom.mids.append( mspl.matid )
				par *= -1  # negate a parity for the next triangle
	
	else:
		for mspl in geom.mspls:
			for fi in range( len( mspl.vids ) // 3 ):
				geom.faces += [ mspl.vids[fi*3], mspl.vids[fi*3+1], mspl.vids[fi*3+2], 0 ]
				geom.mids.append( mspl.matid )

def RemoveDoubles( geom ):
	dff_props = bpy.context.scene.gta_tools.dff_props
	th_co = dff_props.remdbls_th_co
	th_no = dff_props.remdbls_th_no
	if dff_props.dump_stat: print("  Remove Vertices Doubled in both Coordinate and Normal space")
	
	## set normals, if normals are not imported.
	if 0 == len( geom.norms ):
		geom.norms = [0]*len( geom.verts )
		no_count   = [0]*(len( geom.verts )//3)
		for f in range ( len( geom.faces ) //4 ):
			##  n = (v1-v0)x(v2-v0)
			v = (   Vector( geom.verts[geom.faces[f*4  ]*3:geom.faces[f*4  ]*3+3] ),
					Vector( geom.verts[geom.faces[f*4+1]*3:geom.faces[f*4+1]*3+3] ),
					Vector( geom.verts[geom.faces[f*4+2]*3:geom.faces[f*4+2]*3+3] ) )
			normal = ( ( v[1] - v[0] ).cross( v[2] - v[0] ) ).normalized()
			for iv in range( 3 ):
				for ia in range(3):
					geom.norms[geom.faces[f*4+iv]*3+ia] += normal[ia]
				no_count[geom.faces[f*4+iv]] += 1
		
		for iv in range( len( geom.verts ) //3 ):
			if 1 < no_count[iv]:
				for i in range( 3 ):
					geom.norms[iv*3+i] /= no_count[iv]
	
	## set vertex array
	verts = []  # [ index, index_matched, co_x, co_y, co_z, no_x, no_y, no_z ]
	for iv in range( len( geom.verts ) // 3 ):
		verts.append( [iv, -1] + list( geom.verts[ iv*3 : iv*3 + 3 ] ) + list( geom.norms[ iv*3 : iv*3 + 3 ] ) )
	#if ( geom.flags & GEOMETRY_NORMALS ):
	#	for iv in range( len( geom.verts ) // 3 ):
	#		verts.append( [iv, -1] + list( geom.verts[ iv*3 : iv*3 + 3 ] ) + list( geom.norms[ iv*3 : iv*3 + 3 ] ) )
	#else:
	#	for iv in range( len( geom.verts ) // 3 ):
	#		verts.append( [iv, -1] + list( geom.verts[ iv*3 : iv*3 + 3 ] ) + [0, 0, 0] )
	
	sorted_verts = sorted( verts, key=lambda x: x[2] )
	
	## find doubeled verts
	for iv,co0 in enumerate( sorted_verts[: len(sorted_verts) - 1] ):
		if -1 == verts[co0[0]][1]:
			for co1 in sorted_verts[iv+1:]:
				if th_co > co1[2] - co0[2]:
					if th_co > ( Vector(co1[2:5]) - Vector(co0[2:5]) ).length:  # Compare Coordinates
						if th_no > ( Vector(co1[5:8]) - Vector(co0[5:8]) ).length:  # Compare Normals
							verts[co1[0]][1] = co0[0]
				else:
					break
	
	## make new verts index table
	new_index = []  # index: new vert index / value: old vert index
	old_index = []  # index: old vert index / value: new vert index
	for co in verts:
		if -1 == co[1]:
			old_index.append( len( new_index ) )
			new_index.append( co[0] )
		else:
			old_index.append( -1 )
	for co in verts:
		if -1 == old_index[co[0]]:
			old_index[co[0]] = old_index[co[1]]
	
	if dff_props.dump_stat: print("  Reduce Geometry Vertices : " + str(len(geom.verts) // 3) + " -> " + str( len( new_index ) ) )
	
	## Replace Geometry Data
	geom.verts = []
	for iv in new_index: geom.verts += verts[iv][2:5]
	
	#if ( geom.flags & GEOMETRY_NORMALS ):
	geom.norms = []
	for iv in new_index: geom.norms += verts[iv][5:8]
	
	if dff_props.char_skin:
		new_vws = []
		for iv in new_index: new_vws.append( geom.vws[iv] )
		geom.vws = new_vws
		new_bvis = []
		for iv in new_index: new_bvis.append( geom.bvis[iv] )
		geom.bvis = new_bvis
	
	geom.uvfaces = geom.faces[:]
	geom.faces = []
	for iface in range( len( geom.uvfaces ) //4 ):
		geom.faces += [ old_index[ geom.uvfaces[iface*4] ], old_index[ geom.uvfaces[iface*4+1] ], old_index[ geom.uvfaces[iface*4+2] ], 0 ]

def SetTex( material, name_tex, usedtexs = None, map = None, tex_dict = None ):
	dff_props = bpy.context.scene.gta_tools.dff_props
	alpha_mode = bpy.context.scene.gta_tools.map_props.alpha_mode
	
	## Set Color Texture
	# Color Texture Settings
	texslot = material.texture_slots.add()
	texslot.texture_coords = 'UV'
#	material.use_transparent_shadows = True
	
	
	# for MAP Imoprt
	if None != map:
		if name_tex in tex_dict:
			tex = tex_dict[name_tex]
			if tex.col_path in map.used_texs:
				texslot.texture = map.used_texs[tex.col_path]
			else:
				if None == texslot.texture:
					texslot.texture = bpy.data.textures.new( tex.col_name, type = 'IMAGE' )
					if "" != map.tex_path:
						texslot.texture.image = texslot.texture.image = load_image( tex.col_path + ".bmp", map.tex_path )
					rna_idprop_ui_prop_get( texslot.texture, "path", create=True )
					texslot.texture["path"] = tex.col_path
					map.used_texs[tex.col_path] = texslot.texture
		
		## !!!!!!!!!!!!!!!!!
		else:
			texslot.texture = bpy.data.textures.new( name_tex, type = 'IMAGE' )
	
	
	# for DFF Import
	else:
		tex_path = os.path.split( dff_props.imp_filepath )[0]
		
		## if is already used
		if name_tex in usedtexs[0]:
			texslot.texture = usedtexs[0][name_tex]
		
		## not used
		else:
			texslot.texture = bpy.data.textures.new( name_tex, type = 'IMAGE' )
			#texslot.texture.image = load_image( name_tex + ".bmp", tex_path )
			usedtexs[0][name_tex] = texslot.texture
			rna_idprop_ui_prop_get( texslot.texture, "color_texture" , create=True )
			texslot.texture["color_map"]= name_tex
	
	## Set Alpha Texture
	name_alp = name_tex + "a"
	
	# for MAP Imoprt
	if None != map:
		if name_tex in tex_dict:
			tex = tex_dict[name_tex]
			if "" != tex.alp_name:
				if "BMP32" == alpha_mode:
					texslot.texture.use_alpha           = True
					texslot.texture.use_calculate_alpha = False
					texslot.use_map_alpha               = True
					texslot.alpha_factor                = 1.0
					material.use_transparency  = True
					material.alpha             = 0.0
					material.transparency_method = "RAYTRACE"
#					material.transparency_method = "Z_TRANSPARENCY"
				elif "BMP24" == alpha_mode:
					# Alpha Texture Settings
					texslot = material.texture_slots.add()
					texslot.texture_coords = 'UV'
					if tex.alp_path in map.used_texs:
						texslot.texture = map.used_texs[tex.alp_path]
					else:
						if None == texslot.texture:
							texslot.texture = bpy.data.textures.new( tex.alp_name, type = 'IMAGE' )
							texslot.texture.image = load_image( tex.alp_path + ".bmp", map.tex_path )
							rna_idprop_ui_prop_get( texslot.texture, "path", create=True )
							texslot.texture["path"] = tex.alp_path
							map.used_texs[tex.alp_path] = texslot.texture
					
					texslot.texture.use_alpha           = False
					texslot.texture.use_calculate_alpha = False
					texslot.use_map_color_diffuse    = False
					texslot.use_map_alpha            = True
					texslot.blend_type = 'MULTIPLY'
					material.use_transparency  = True
					material.transparency_method = "RAYTRACE"
#					material.transparency_method = "Z_TRANSPARENCY"
	
	## for DFF Import
	elif os.path.exists( tex_path + "\\" + name_alp + ".bmp" ):
		# Alpha Texture Settings
		texslot = material.texture_slots.add()
		texslot.texture_coords = 'UV'
		
		tex_path = os.path.split( dff_props.imp_filepath )[0]
		## if already used
		if name_alp in usedtexs[1]:
			texslot.texture = usedtexs[1][name_alp]
		
		## not used
		else:
			texslot.texture = bpy.data.textures.new( name_alp, type = 'IMAGE' )
			#texslot.texture.image = load_image( name_alp + ".bmp", tex_path )
			usedtexs[0][name_alp] = texslot.texture
		
		texslot.texture.use_alpha           = False
		texslot.texture.use_calculate_alpha = False
		texslot.use_map_color_diffuse    = False
		texslot.use_map_alpha            = True
		texslot.blend_type = 'MULTIPLY'
		material.use_transparency  = True
		material.transparency_method = "RAYTRACE"


def ReadFrameList( file, frams ):
	dff_props = bpy.context.scene.gta_tools.dff_props
	try:
		## Read Frame List
		hanim_flag = False
		h = CheckHeader( file, 0x0E )  # Frame List
		endof_framelist = h.ofs + h.size
		CheckHeader( file, 0x01 )  # Struct
		data = struct.unpack( "<i", file.read( 4 ) )
		num_frams = data[0]
		if dff_props.dump_stat: print( "Number of Frames : " + str( num_frams ) )
		
		## Read Frames
		for fi in range( num_frams ):
			fram = ClassFrame()
			fram.rot = Matrix( ( struct.unpack( "<3f", file.read( 12 ) ), struct.unpack( "<3f", file.read( 12 ) ), struct.unpack( "<3f", file.read( 12 ) ) ) )
			fram.pos = Vector( struct.unpack( "<3f", file.read( 12 ) ) )
			fram.pfram = struct.unpack( "<2i", file.read( 8 ) )[0]
			frams.append( fram )
		
		for fi, fram in enumerate( frams ):
			if dff_props.dump_stat: print( "Read Frame : #" + str( fi ) )
			h = CheckHeader( file, 0x03 )  # Extension
			endof_extension = h.ofs + h.size
			
			###
			# The extension itself got 2 child section which may appear in random order
			headers = ReadHeaders( file, endof_extension )
			for h in headers:
				file.seek( h.ofs, 0 )
				endof_section = h.ofs + h.size
				if 0x11e == h.id:  # HAnimPLG
					hanim_flag = True
					data = struct.unpack( "<3i", file.read( 12 ) )
					fram.bid = data[1]
					num_bones = data[2]
					if 0 < num_bones:
						file.seek( 8, 1 )  # Skip UnknownInt x2
						for ibone in range(num_bones):
							data = struct.unpack( "<3i", file.read( 12 ) )
							fram.btable.append( data )
				elif 0x253f2fe == h.id:  # Frame
					fram.name = bytes.decode( file.read( h.size ), errors='replace' ).split( '\0' )[0]
					if dff_props.dump_stat: print( "  Frame Name : " + fram.name )
				file.seek( endof_section, 0 )
			file.seek( endof_extension, 0 )
			
		file.seek( endof_framelist, 0 )  # Skip for Now!!
		return hanim_flag
	
	except: raise ImportError

def ReadGeometry( file, geom ):
	dff_props = bpy.context.scene.gta_tools.dff_props
	try:
		h = CheckHeader( file, 0x0f )  # Geometry
		endof_geom = h.ofs + h.size
		ver_id = h.ver
		CheckHeader( file, 0x01 )  # Struct
		
		## read Geometry Flags
		data = struct.unpack( "<H", file.read( 2 ) )
		geom.flags = data[0]
		if dff_props.dump_stat: print( "  Geometry Flags : " + hex( geom.flags ) )
		
		## Read Num Faces/Verts/Frames
		data = struct.unpack( "<2b3i", file.read( 14 ) )
		num_uvf   = data[0]
		num_faces = data[2]
		num_verts = data[3]
		num_frams = data[4]
		if dff_props.dump_stat:
			print( "  Number of UV Faces : " + str( num_uvf ) )
			print( "  Number of Faces    : " + str( num_faces ) )
			print( "  Number of Verts    : " + str( num_verts ) )
			print( "  Number of Frames   : " + str( num_frams ) )
		
		if 0x1803FFFF != ver_id:
			data = struct.unpack( "<3i", file.read( 12 ) )  # for GTA3/VC Params( Ambient, Diffuse, Specular )
		
		## Read VCOLs
		if ( geom.flags & GEOMETRY_PRELIT ):
			geom.vcols = struct.unpack( '<%dB' % ( num_verts * 4 ), file.read( 4 * num_verts ) )
		
		## Read UVs
		if ( geom.flags & GEOMETRY_UVTEXTURE or geom.flags & GEOMETRY_ETEXTURE ):
			for uvf in range( num_uvf ):
				for v in range( num_verts ):
					uvtemp = struct.unpack( "<2f", file.read( 8 ) )  # UV Coords
					geom.uvs.append( [uvtemp[0], 1.0-uvtemp[1]] )
		
		## Read Faces
		if dff_props.use_msplit:
			file.seek( num_faces * 8, 1 )
		else:
			for i in range( num_faces ):
				data = struct.unpack( "<4h", file.read( 8 ) )  # Face List
				#geom.faces.append( [data[1], data[0], data[3], data[2], i] )  # [v1, v2, v3, v4( matid ), index]
				geom.faces += [data[1], data[0], data[3], 0 ]
				geom.mids.append( data[2] )
		
		## Read Boundary
		boundsphere = struct.unpack( "<4f", file.read( 16 ) )  # Bounding Sphere
		file.seek( 8, 1 )
		
		## Read Verts
		geom.verts = struct.unpack( '<%df' % ( num_verts * 3 ), file.read( 12 * num_verts ) )
		
		## Read Normals
		if ( geom.flags & GEOMETRY_NORMALS ):
			geom.norms = struct.unpack( '<%df' % ( num_verts * 3 ), file.read( 12 * num_verts ) )
		
		## Read Material List
		CheckHeader( file, 0x08 )  # Material List
		CheckHeader( file, 0x01 )  # Struct
		data = struct.unpack( "<i", file.read( 4 ) )
		num_mats = data[0]
		if dff_props.dump_stat: print( "  Number of Materials : " + str( num_mats ) )
		data = struct.unpack( '<%di' % ( num_mats ), file.read( 4 * num_mats ) )
		
		for i in range( num_mats ):
			## Read Materials
			mat = ClassMaterial()
			h = CheckHeader( file, 0x07 )  # Material
			endof_mat = h.ofs + h.size
			CheckHeader( file, 0x01 )  # Struct
			data = struct.unpack( "<i4B", file.read( 8 ) )
			mat.color = list(data[1:])
			data = struct.unpack( "<2i3f", file.read( 20 ) )
			num_texs = data[1]
			mat.color.append(data[4])  #  Diffuse Intensity (in "KAMS") !!!Import temporarily!!!
			if dff_props.dump_stat: print( "  Number of Textures : " + str( num_texs ) )
			
			for i in range( num_texs ):
				## Read Textures
				tex = ClassTexture()
				h = CheckHeader( file, 0x06 )  # Texture
				endof_tex = h.ofs + h.size
				CheckHeader( file, 0x01 )  # Struct
				data = struct.unpack( "<2H", file.read( 4 ) )
				tex.flags = data[0]
				if dff_props.dump_stat: print( "  Filter Flags : " + hex( tex.flags ) )
				
				h = CheckHeader( file, 0x02 )  # String
				tex.name_tex = bytes.decode( file.read( h.size ), errors='replace' ).split( '\0' )[0].lower()
				if dff_props.dump_stat: print( "  Texture Name : " + tex.name_tex )
				
				h = CheckHeader( file, 0x02 )  # String
				tex.name_mask = bytes.decode( file.read( h.size ), errors='replace' ).split( '\0' )[0].lower()
				if dff_props.dump_stat: print( "  Mask Name : " + tex.name_mask )
				
				mat.texs.append( tex )
				file.seek( endof_tex, 0 )  # Skip Extension @Texture
				
			## Read Extension @Material
			h = CheckHeader( file, 0x03 )  # Extension @Material
			endof_extension = h.ofs + h.size
			
			headers = ReadHeaders( file, endof_extension )
			for h in headers:
				file.seek( h.ofs, 0 )
				if 0x120 == h.id:  # Material Effects PLG
					file.seek( h.size, 1 )  # Skip For Now!!
				elif 0x253f2fc == h.id:  # Reflection Materil
					data = struct.unpack( "<5fi", file.read( 24 ) )
					mat.ref = data[:5]
				elif 0x253f2f6 == h.id:  # Specular Materil
					spec_level = struct.unpack( "<f", file.read( 4 ) )[0]
					spec_tex = bytes.decode( file.read( h.size - 8 ), errors='replace' ).split( '\0' )[0].lower()
					mat.spec = [ spec_level, spec_tex ]
				else:
					file.seek( h.size, 1 )
			file.seek( endof_extension, 0 )
			
			geom.mats.append( mat )
			file.seek( endof_mat, 0 )  # Skip
		
		## Read Extension @Geometry
		CheckHeader( file, 0x03 )  # Extension
		
		## Material Split
		if dff_props.use_msplit:
			h = CheckHeader( file, 0x50E )  # BinMeshPLG
			if dff_props.use_msplit:
				data = struct.unpack( "<3i", file.read( 12 ) )
				geom.ftype = data[0]
				num_splits = data[1]
				
				for si in range( num_splits ):
					mspl = ClassMSplit()
					data = struct.unpack( "<2i", file.read( 8 ) )
					num_indices = data[0]
					mspl.matid = data[1]
					mspl.vids = struct.unpack( '<%di' % ( num_indices ), file.read( 4 * num_indices ) )
					geom.mspls.append( mspl )
			else:
				file.seek( h.size, 1 )
		
		## Skin PLG ( for Char Import )
		if dff_props.char_skin:
			CheckHeader( file, 0x116 )  # Skin PLG
			data = struct.unpack( "<4B", file.read( 4 ) )
			num_bones   = data[0]
			num_spbones = data[1]
			geom.spunk  = data[2]
			geom.spids = struct.unpack( '<%dB' % ( num_spbones ), file.read( num_spbones ) )
			for vi in range( len( geom.verts ) // 3 ):
				geom.bvis.append( struct.unpack( "<4B", file.read( 4 ) ) )
			for vi in range( len( geom.verts ) // 3 ):
				geom.vws.append( struct.unpack( "<4f", file.read( 16 ) ) )
			for ibone in range( num_bones ):
				ibmat = list( struct.unpack( "<16f", file.read( 4*16 ) ) )
				ibmat[15] = 1.0  # for invertion matrix
				geom.ibmats.append( Matrix( ( ibmat[0:4], ibmat[4:8], ibmat[8:12], ibmat[12:16] ) ) )
		
		file.seek( endof_geom, 0 )  # Skip to next geom
		return
		
	except: raise ImportError

def ReadGeometryList( file, geoms ):
	dff_props = bpy.context.scene.gta_tools.dff_props
	try:
		## Read Geometry List
		CheckHeader( file, 0x1A )  # Geometry List
		CheckHeader( file, 0x01 )  # Struct
		data = struct.unpack( "<i", file.read( 4 ) )
		num_geoms = data[0]
		if dff_props.dump_stat: print( "Number of Geometries : " + str( num_geoms ) )
		
		## Read Geometries
		for i in range( num_geoms ):
			geom = ClassGeometry()
			if dff_props.dump_stat: print( "Read Geometry : #" + str( i ) )
			ReadGeometry( file, geom )
			geoms.append( geom )
		
		return geoms
	
	except: raise ImportError

def ReadAtomics( file, atoms, endof_clump ):
	dff_props = bpy.context.scene.gta_tools.dff_props
	try:
		## Read Atmics
		headers = ReadHeaders( file, endof_clump, 0x14 )
		for h in headers:
			file.seek( h.ofs, 0 )
			endof_data = h.ofs + h.size
			## Read Atmic
			if dff_props.dump_stat: print( "Read Atomic : #" + str( len( atoms ) ) )
			atom = ClassAtomic()
			CheckHeader( file, 0x01 )  # Struct
			data = struct.unpack( "<4i", file.read( 16 ) )
			atom.ifram = data[0]
			atom.igeom = data[1]
			atoms.append( atom )
			if dff_props.dump_stat: print( "  Frame ID, Geometry ID : " + str( atom.ifram ) + ", " + str( atom.igeom ) )
			file.seek( endof_data, 0 )
	
	except: raise ImportError

def ReadCollision( file, coll ):
	dff_props = bpy.context.scene.gta_tools.dff_props
	try:
		if dff_props.dump_stat: print( "-----\nRead Collision" )
		## Read Extension
		CheckHeader( file, 0x03 )  # Extension
		
		## Read Collision
		h = CheckHeader( file, 0x253f2fa )  # Collision
		beg_coll = h.ofs + 4
		
		## Header
		# FourCC, Name, Size @ Header
		data = struct.unpack( "<2i", file.read( 8 ) )
		fourcc = data[0]
		if ( 0x334c4f43 != fourcc ) and ( 0x324c4f43 != fourcc ):  # "COL3" or "COL2"
			if dff_props.dump_stat: print( "Error: Non Supported Collision Format : FourCC = " + hex(fourcc) )
			raise ImportError
		coll.fourcc = fourcc
		coll.size = data[1]
		coll.name = bytes.decode( file.read( 20 ), errors='replace' ).split( '\0' )[0]
		
		if dff_props.dump_stat:
			print( "  Collision Name : " + coll.name )
			if   ( 0x334c4f43 == fourcc ): print( "  Collision Version : COL3" )
			elif ( 0x324c4f43 == fourcc ): print( "  Collision Version : COL2" )
		
		
		# TBounds @Header
		data = struct.unpack( "<11f", file.read( 44 ) )  # unknown(4), TBounds(40)
		coll.bbox[0]     = Vector( data[1:4] )   # box_min
		coll.bbox[1]     = Vector( data[4:7] )   # box_max
		coll.bsphere[0]  = Vector( data[7:10] )  # sphere_senter
		coll.bsphere[1]  = data[10]              # sphere_radius
		
		# NumObjs, Offset @Header
		data = struct.unpack( "<2H8I", file.read( 36 ) )
		num_cspheres = data[0]
		num_cboxes   = data[1]
		num_cmeshes = data[2]
		flags       = data[3]
		coll.flag_fgroup = flags & 0x8
		coll.flag_shadow = flags & 0x16
		if( coll.flag_fgroup ):
			if dff_props.dump_stat:
				print( "Error : Face Group is not Supported." )
				print( "  Collision Mesh Flags : " + hex( flags ) )
		ofs_cspheres = data[4]
		ofs_cboxes   = data[5]
		ofs_cverts   = data[7]
		ofs_cfaces   = data[8]
		if 0x334c4f43 == coll.fourcc and coll.flag_shadow:
			data = struct.unpack( "<3I", file.read( 12 ) )
			num_smeshes = data[0]
			ofs_sverts  = data[1]
			ofs_sfaces  = data[2]
		
		num_cverts = ( ofs_cfaces - ofs_cverts ) // 6
		num_sverts = ( ofs_sfaces - ofs_sverts ) // 6
		if dff_props.dump_stat: print( "  Collision Flags : " + hex( flags ) )
		
		## Body
		# TSphere @Body
		file.seek( beg_coll + ofs_cspheres, 0 )
		for isphere in range( num_cspheres ):
			data = struct.unpack( "<4f4B", file.read( 20 ) )
			sphere = []
			sphere.append( Vector( data[0:3] ) )  # center : TVector
			sphere.append( data[3] )              # radius : float
			sphere.append( data[4:8] )            # surface: TSurface
			coll.spheres.append( sphere )
		
		# TBox @Body
		file.seek( beg_coll + ofs_cboxes, 0 )
		for ibox in range( num_cboxes ):
			data = struct.unpack( "<6f4B", file.read( 28 ) )
			box = []
			box.append( Vector( data[0:3] ) )  # min     : TVector
			box.append( Vector( data[3:6] ) )  # max     : TVector
			box.append( data[6:10] )           # surface : TSurface
			coll.boxes.append( box )
		
		# TVertex @Body
		# To convert such an int16 number to a floating point number, simply divide it by 128.0. 
		file.seek( beg_coll + ofs_cverts, 0 )
		for ivert in range( num_cverts ):
			data = struct.unpack( "<3h", file.read( 6 ) )
			co = Vector( ( float( data[0] / 128.0 ), float( data[1] / 128.0 ), float( data[2] / 128.0 ) ) ) # INT16 to float
			coll.verts.extend( co )
		
		
		# TFaceGroup @Body
		# !!! Skip for now.
		
		# TFace @Body
		file.seek( beg_coll + ofs_cfaces, 0 )
		for iface in range( num_cmeshes ):
			data = struct.unpack( "<3H2B", file.read( 8 ) )
			
			verts  = list( data[0:3] )  # a, b, c : uint16;
			mat    = data[3]            # material: uint8;
			light  = data[4]            # light   : uint8;
			
			verts.append( 0 )  # for foreach_set to "vertices_raw"
			coll.faces.extend( verts )
			coll.surfs.append( ( mat, light ) )
			
		## Shadow Mesh
		if 0x334c4f43 == coll.fourcc and coll.flag_shadow:
			# TVertex @Body
			# To convert such an int16 number to a floating point number, simply divide it by 128.0. 
			file.seek( beg_coll + ofs_sverts, 0 )
			for ivert in range( num_sverts ):
				data = struct.unpack( "<3h", file.read( 6 ) )
				co = Vector( ( float( data[0] / 128.0 ), float( data[1] / 128.0 ), float( data[2] / 128.0 ) ) ) # INT16 to float
				coll.sverts.extend( co )
			
			# TFace @Body
			file.seek( beg_coll + ofs_sfaces, 0 )
			for iface in range( num_smeshes ):
				data = struct.unpack( "<3H2B", file.read( 8 ) )
				verts  = list( data[0:3] )  # a, b, c : uint16;
				mat    = data[3]            # material: uint8;
				light  = data[4]            # light   : uint8;
				
				verts.append( 0 )  # for foreach_set to "vertices_raw"
				coll.sfaces.extend( verts )
				coll.ssurfs.append( ( mat, light ) )
	
	except: raise ImportError

def ReadClump( file, clump, endof_clump ):
	dff_props = bpy.context.scene.gta_tools.dff_props
	try:
		## Read Clump
		CheckHeader( file, 0x01 )  # Struct
		data = struct.unpack( "<iii", file.read( 12 ) )
		num_obj = data[0]
		if dff_props.dump_stat: print( "Number of Objects : " + str( num_obj ) )
		
		char_flag = ReadFrameList( file, clump.frams )
		if True == dff_props.char_skin and False == char_flag:
			bpy.context.scene.gta_tools.msg +=  "Error : This Model has no HAnimPLG sections\n"
			raise ImportError
		ReadGeometryList( file, clump.geoms )
		ReadAtomics( file, clump.atoms, endof_clump )
		if dff_props.vehicle:
			ReadCollision( file, clump.coll )
		return
	
	except: raise ImportError

def ReadDFF( file, dff ):
	dff_props = bpy.context.scene.gta_tools.dff_props
	try:
		## Check Clumps
		headers = ReadHeaders( file, dff.end, 0x10 )  # Clump
		if 0 == len( headers ):
			bpy.context.scene.gta_tools.msg +=  "Error : No Clumps.\n"
			return
		else:
			dff.ver_id = headers[0].ver
			if dff_props.dump_stat: print( "RW Version : " + rw_version[dff.ver_id] )
		if dff_props.dump_stat: print( "Number of Clumps : " + str( len( headers ) ) )
		for ih, h in enumerate( headers ):
			if dff_props.dump_stat: print( "Size of Clump #" + str( ih ) + " : " + str( h.size ) + " bytes" )
		
		## Read Clumps
		for ih, h in enumerate( headers ):
			if dff_props.dump_stat: print( "-----\nRead Clump : #" + str( ih ) )
			file.seek( h.ofs, 0 )  # OFS of Clump Data
			clump = ClassClump()
			ReadClump( file, clump, h.ofs + h.size )
			if 1 == len( headers ): clump.name = dff.name
			else: clump.name = dff.name + "_Clump#" + str( i )
			dff.clumps.append(clump)
	
	except: raise ImportError

def CheckTXD( tex_list, txd ,map ):
	results = []
	texs = []
	not_found = []
	txd_path = txd.img.name + "\\" + txd.name
	
	## Open File
	try:
		file = open( txd.img.path, "rb" )
	except:
		if dff_props.dump_stat: print( "-----\nError : failed to open " + txd.img.path )
		return
	
	file.seek( txd.ini, 0 )
	from . import extract_txd
	texs = extract_txd.ExtractTXD( file, tex_list, map.tex_path, extract = map.extract_txd, map = map, txd_path = txd_path )
	file.close
	
	return texs

def CheckTXDs( dff, txds, map ):
	## Make Tex List
	tex_list = []
	texs = []
	for clump in dff.clumps:
		for geom in clump.geoms:
			for mat in geom.mats:
				for tex in mat.texs:
					tex_list.append( tex.name_tex )
	
	## Make Tex Dicts
	tex_list = list( set( tex_list ) )
	for txd in txds:
		texs.extend( CheckTXD( tex_list, txd ,map ) )
	
	for tex in tex_list:
		for txd in txds:
			map.dbg.append( " Not Found: " + txd.img.name + "\\" + txd.name + "\\" + tex )
	
	for tex in texs:
		dff.tex_dict[tex.col_name] = tex

def CreateObject( dff, map = None ):
	## Create Object
	dff_props = bpy.context.scene.gta_tools.dff_props
	usedtexs  = [{},{}]  # dict{ key = texturename : value = bpy.types.texture }
	groups = []
	
	for iclump,clump in enumerate( dff.clumps ):
		frams = clump.frams
		geoms = clump.geoms
		atoms = clump.atoms
		
		mesh_frams     = []
		node_frams     = []
		skin_frams     = []  # for Char Skin
		bone_frams     = []  # for Char Skin
		objs           = []
		obj_dict       = {}  # for parenting
		coll_objs      = []  # for Vehicle Model
		skin_objs      = []  # for Char Skin
		arm_objs       = []  # for Char Skin
		bones          = []  # for Char Skin
		skin_obj_index = 0   # for Char Skin; Const: Experimentally '0'
		root_bid       = 0   # for Char Skin
		skin_ifram     = 0   # for Char Skin
		skin_igeom     = 0   # for Char Skin
		root_ifram     = 1   # for Char Skin
		ifram2ibone    = {}  # for Char Skin; key = ifram, value = ibone
		bid_dict       = {}  # for Char Skin; key = bid, value = [ Bone No., type, ifram ]
		
		if dff_props.dump_stat: print( "-----\nCreate Objects: Clump #" + str( iclump ) )
		for ifram, fram in enumerate( frams ):
			## Check Object Type
			igeom = -1
			if dff_props.char_skin:  # for Char Import
				if skin_ifram == ifram:
					igeom = skin_igeom
					skin_frams.append( ifram )
				else:
					bone_frams.append( ifram )
					if root_bid == fram.bid: root_ifram = ifram
			else:
				for a in atoms:  # Check the Frame is in Atomics or not
					if ifram == a.ifram:
						igeom = a.igeom
						break
				if igeom == -1:
					node_frams.append( ifram )
				else:
					mesh_frams.append( ifram )
			
			## Create Meshes
			if -1 != igeom:
				if dff_props.char_skin: fram_name = clump.name
				else:                   fram_name = fram.name
				geom = geoms[igeom]
				if dff_props.dump_stat: print( "Create Mesh : " + "( " + str( ifram ) + " )" + fram_name )
				m = bpy.data.meshes.new( fram_name )
				
				## Make Materials / Assign Texture
				for mat in geom.mats:
					material = bpy.data.materials.new( "Mat_" + dff.name )
					material.diffuse_color = [float( mat.color[0] )/255.0, float( mat.color[1] )/255.0, float( mat.color[2] )/255.0]
					material.alpha = float( mat.color[3] )/255.0
					material.diffuse_intensity = mat.color[4]
					#material.use_transparency  = True
					
					if 0 != len( mat.ref ):
						material.specular_color     = mat.ref[:3]
						material.specular_alpha     = mat.ref[3]
						material.specular_intensity = float( mat.ref[4] )
						#material.specular_intensity = float( mat.ref[4] )/255.0
						#if dff_props.dump_stat: print( mat.ref[4] )
					else:  # If Not Specify Reflection(Specular), Set Surface to "soft face", this assignments are temporarily.
						material.specular_intensity = 0.0
						material.specular_hardness  = 0
					if 0 != len( mat.spec ):
						rna_idprop_ui_prop_get( material, "GTAMAT_spec_level"   , create=True )
						rna_idprop_ui_prop_get( material, "GTAMAT_spec_texture" , create=True )
						material["GTAMAT_spec_level"]   = mat.spec[0]
						material["GTAMAT_spec_texture"] = mat.spec[1]
					
					for tex in mat.texs:
						SetTex( material, tex.name_tex, usedtexs, map, dff.tex_dict )
					
					m.materials.append( material )
				
				if dff_props.use_msplit: MaterialSplit( geom )
				if dff_props.use_remdbls: RemoveDoubles( geom )
				
				## Create Verts
				m.vertices.add( len( geom.verts ) // 3 )
				m.vertices.foreach_set( "co", geom.verts )
				
				## Create Faces
				num_faces = len( geom.faces ) // 4
				m.faces.add( num_faces )
				m.faces.foreach_set( "vertices_raw", geom.faces )
				m.faces.foreach_set( "material_index", geom.mids )
				m.faces.foreach_set( "use_smooth", [True]*num_faces )
				
				## Set UVs
				if ( geom.flags & GEOMETRY_UVTEXTURE or geom.flags & GEOMETRY_ETEXTURE ):
					uvf = m.uv_textures.new()
					uvs = []
					if dff_props.use_remdbls:
						for fi, f in enumerate( m.faces ):
							uvs += geom.uvs[ geom.uvfaces[fi*4] ] + geom.uvs[ geom.uvfaces[fi*4+1] ] + geom.uvs[ geom.uvfaces[fi*4+2] ] + [0.0, 0.0]
					else:
						for fi, f in enumerate( m.faces ):
							uvs += geom.uvs[ f.vertices[0] ] + geom.uvs[ f.vertices[1] ] + geom.uvs[ f.vertices[2] ] + [0.0, 0.0]
					uvf.data.foreach_set( "uv_raw", uvs )
					#uvf.data.foreach_set( "use_twoside", [False]*num_faces )  ## for Blender2.5x <--> 2.60 compatibility
					
					if bpy.app.version[1] < 60:  ## for Blender2.5x <--> 2.60 compatibility
						for fi, f in enumerate( m.faces ):
							mat = m.materials[f.material_index]
							if( len( mat.texture_slots.items() ) ):
								uvf.data[fi].image = mat.texture_slots[0].texture.image
								uvf.data[fi].use_image = True
					else:
						for fi, f in enumerate( m.faces ):
							mat = m.materials[f.material_index]
							if( len( mat.texture_slots.items() ) ):
								uvf.data[fi].image = mat.texture_slots[0].texture.image
								#uvf.data[fi].use_image = True
				
				m.update()
				m.vertices.foreach_set( "normal", geom.norms )  # Vert NORs seem to be recalculated by "mesh.update()"
				#if( geom.flags & GEOMETRY_NORMALS ): m.vertices.foreach_set( "normal", geom.norms )  # Vert NORs seem to be recalculated by "mesh.update()"
				
				## Create Object
				o = bpy.data.objects.new( fram_name, m )
				bpy.context.scene.objects.link( o )
				o.show_name = False
				o.show_x_ray = False
				o.rotation_mode = 'QUATERNION'
				o.location = fram.pos
				o.rotation_quaternion = fram.rot.to_quaternion()
				objs.append( o )
				obj_dict[ifram] = o
				
				if dff_props.char_skin:
					## Create Armature
					arm = bpy.data.armatures.new( "Armature" )
					arm_obj = bpy.data.objects.new( fram_name + "_Armature", arm )
					if dff_props.dump_stat: print( "Create Armature: " + arm_obj.name )
					bpy.context.scene.objects.link( arm_obj )
					arm_obj.show_name = False
					arm_obj.show_x_ray = True
					arm_obj.rotation_mode = 'QUATERNION'
					arm_obj.location = fram.pos
					arm_obj.rotation_quaternion = fram.rot.to_quaternion()
					arm.show_names = True
					arm.show_axes  = True
					skin_objs.append( o )
					arm_objs.append( arm_obj )
			
			## Create Nodes
			elif False == dff_props.char_skin and False == dff.skip_nodes:
				if dff_props.dump_stat: print( "Create Node : " + "( " + str( ifram ) + " )" + fram.name )
				o = bpy.data.objects.new( fram.name, None )
				o.empty_draw_size = 0.03
				o.empty_draw_type = 'CUBE'
				o.show_x_ray = True
				o.show_name = False
				bpy.context.scene.objects.link( o )
				
				## Set Rotation and Location
				o.rotation_mode = 'QUATERNION'
				o.location = fram.pos
				o.rotation_quaternion = fram.rot.to_quaternion()
				objs.append( o )
				obj_dict[ifram] = o
		
		## Create Bones, Parenting, Bone/Weight Setup; for Char Skin
		if dff_props.char_skin:  # for Char Skin
			skin_geom = geoms[skin_igeom]
			skin_obj = skin_objs[skin_obj_index]
			arm_obj = arm_objs[skin_obj_index]
			
			## Set Tables for converting IDs
			for item in frams[root_ifram].btable:
				bid_dict[item[0]] = list( item[1:] )
				skin_obj.vertex_groups.new()
			for ibone, ifram in enumerate(bone_frams):
				ifram2ibone[ifram] = ibone
				bid_dict[frams[ifram].bid].append( ifram )
			
			## Create Bones
			bpy.context.scene.objects.active = arm_obj
			bpy.ops.object.mode_set( mode = 'EDIT' )
			for ibone, ifram in enumerate( bone_frams ):
				fram = frams[ifram]
				if dff_props.ren_bone:
					bone_name = fixed_bone_name[fram.bid]
				else:
					bone_name = fram.name
				if dff_props.dump_stat: print( "Create Bone : " + "( " + str( ifram ) + ", " + str( bid_dict[fram.bid][0] ) + ", " + str( fram.bid ) + " )" + bone_name )
				
				## Set Vertex Group Name
				skin_obj.vertex_groups[bid_dict[fram.bid][0]].name = bone_name
				
				## Create Bones
				bone = arm_obj.data.edit_bones.new( bone_name )
				bones.append( bone )
				
				## Set "Bone ID", "Bone Type" as "Custom Property" of Blender
				rna_idprop_ui_prop_get( bone, "org_name", create=True )
				rna_idprop_ui_prop_get( bone, "bone_id", create=True )
				rna_idprop_ui_prop_get( bone, "bone_type", create=True )
				bone["org_name"] = fram.name
				bone["bone_id"] = fram.bid
				bone["bone_type"] = bid_dict[fram.bid][1]
			
			## Parenting
			if dff_props.dump_stat: print( "-----\nParenting" )
			skin_obj.parent = arm_obj
			for ibone, ifram in enumerate(bone_frams):
				bone = bones[ibone]
				fram  = frams[ifram]
				bone.tail = [0.05, 0.0, 0.0]
				if root_ifram != ifram:
					bone.parent = bones[ifram2ibone[fram.pfram]]
					if dff_props.dump_stat: print( "( " + str( ifram ) + " )" + bone.name + " -> " + "( " + str( fram.pfram ) + " )" + bone.parent.name )
				else:
					if dff_props.dump_stat: print( "( " + str( ifram ) + " )" + bone.name )
					pass
			
			## Set Rotation and Location
			tail = Vector( [0.0, 0.05, 0.0 ] )
			vecz = Vector( [0.0, 0.0 , 1.0 ] )  # for calculationg roll angle decision
			for bone in bones:
				bn = bid_dict[bone["bone_id"]][0]
				mat = geom.ibmats[bn].inverted()
				bone.head = mat.to_translation()
				bone.tail = tail.copy()
				bone.tail.rotate( mat.to_3x3() )
				bone.tail += bone.head
				bonez = vecz.copy()
				bonez.rotate( mat.to_3x3() )
				bone.roll = AlignRoll( bone.vector, bone.z_axis, bonez )
			
			##  Add Skin Modifier
			bpy.ops.object.mode_set( mode = 'OBJECT' )
			mod = skin_obj.modifiers.new( "Armature", 'ARMATURE' )
			mod.object = arm_obj
			mod.use_vertex_groups = True
			mod.use_bone_envelopes = False
			
			##  Set Vertex Weights
			for iv, v in enumerate( skin_obj.data.vertices ):
				for iw, vw in enumerate( skin_geom.vws[iv] ):
					bvi = skin_geom.bvis[iv][iw]
					vg = skin_obj.vertex_groups[bvi]
					if 0 < vw:
						vg.add( [iv], vw, 'REPLACE' )
		
		## Parenting, Create Collisions; for non-Char Skin
		else:
			## Parenting
			if dff_props.dump_stat: print( "-----\nParenting" )
			root_obj = None
			for ifram in obj_dict:
				if dff_props.dump_stat: print( ifram, frams[ifram].name, frams[ifram].pfram )
				o = obj_dict[ifram]
				fram  = frams[ifram]
				if fram.pfram in obj_dict:
					o.parent = obj_dict[fram.pfram]
					if dff_props.dump_stat: print( "( " + str( ifram ) + " )" + o.name + " -> " + "( " + str( fram.pfram ) + " )" + o.parent.name )
				else:
					if dff_props.dump_stat: print( "( " + str( ifram ) + " )" + o.name )
					root_obj = o
			
			## Create Collisions; for Vehicle Model
			if dff_props.vehicle:  # for Vehicle Model
				coll = clump.coll
				if dff_props.dump_stat: print( "Create Collision : " + coll.name )
				
				## Create a Root Obj for all collisions.
				# for tree display in Outliner,
				# Not useed in export ops
				coll_dummy = bpy.data.objects.new( "collisions_dummy", None )
				coll_dummy.empty_draw_size = 0.05
				coll_dummy.show_x_ray = True
				coll_dummy.show_name = False
				coll_dummy.location = Vector( ( 0, 0, 0 ) )
				bpy.context.scene.objects.link( coll_dummy )
				coll_dummy.parent = root_obj
				coll_objs.append( coll_dummy )
				
				## Create a Bounding Box
				o = bpy.data.objects.new( "bounding_box", None )
				center = ( coll.bbox[1] + coll.bbox[0] ) /2
				scale  = ( coll.bbox[1] - coll.bbox[0] ) /2
				o.empty_draw_type = 'CUBE'
				o.empty_draw_size = 1.0
				o.show_x_ray = True
				o.show_name = False
				o.location = center
				o.scale = scale
				bpy.context.scene.objects.link( o )
				o.parent = coll_dummy
				coll_objs.append( o )
				
				## Create a Bounding Sphere
				o = bpy.data.objects.new( "bounding_sphere", None )
				o.empty_draw_type = 'SPHERE'
				o.empty_draw_size = 1.0
				o.show_x_ray = True
				o.show_name = False
				o.location = coll.bsphere[0]
				o.scale = [coll.bsphere[1]]*3
				bpy.context.scene.objects.link( o )
				o.parent = coll_dummy
				coll_objs.append( o )
				
				## Create Collision Boxes
				for box in coll.boxes:
					o = bpy.data.objects.new( "collision_box", None )
					center = ( box[1] + box[0] ) /2
					scale  = ( box[1] - box[0] ) /2
					o.empty_draw_type = 'CUBE'
					o.empty_draw_size = 1.0
					o.show_x_ray = True
					o.show_name = False
					o.location = center
					o.scale = scale
					bpy.context.scene.objects.link( o )
					o.parent = coll_dummy
					rna_idprop_ui_prop_get( o, "material"  , create=True )
					rna_idprop_ui_prop_get( o, "flag"      , create=True )
					rna_idprop_ui_prop_get( o, "brightness", create=True )
					rna_idprop_ui_prop_get( o, "light"     , create=True )
					o["material"]   = box[2][0]
					o["flag"]       = box[2][1]
					o["brightness"] = box[2][2]
					o["light"]      = box[2][3]
					coll_objs.append( o )
				
				## Create Collision Spheres
				for sphere in coll.spheres:
					o = bpy.data.objects.new( "collision_sphere", None )
					o.empty_draw_type = 'SPHERE'
					o.empty_draw_size = 1.0
					o.show_x_ray = True
					o.show_name = False
					o.location = sphere[0]
					o.scale = [sphere[1]]*3
					bpy.context.scene.objects.link( o )
					o.parent = coll_dummy
					rna_idprop_ui_prop_get( o, "material"  , create=True )
					rna_idprop_ui_prop_get( o, "flag"      , create=True )
					rna_idprop_ui_prop_get( o, "brightness", create=True )
					rna_idprop_ui_prop_get( o, "light"     , create=True )
					o["material"]   = sphere[2][0]
					o["flag"]       = sphere[2][1]
					o["brightness"] = sphere[2][2]
					o["light"]      = sphere[2][3]
					coll_objs.append( o )
				
				## Create a Collision Mesh
				m = bpy.data.meshes.new( "collision_mesh" )
				m.vertices.add( len( coll.verts ) // 3 )
				m.vertices.foreach_set( "co", coll.verts )
				m.faces.add( len( coll.faces ) // 4 )
				m.faces.foreach_set( "vertices_raw", coll.faces )
				cmat_dict = {}
				for iface, face in enumerate( m.faces ):
					mat = coll.surfs[iface]
					if not( mat in cmat_dict ):
						material = bpy.data.materials.new( "Mat_collision" )
						material.diffuse_intensity = 0.5
						cmat_dict[mat] =len( m.materials )
						rna_idprop_ui_prop_get( material, "GTA_Material" , create=True )
						rna_idprop_ui_prop_get( material, "GTAMAT_coll_light" , create=True )
						material["GTA_Material"]      = "collision"
						material["GTAMAT_coll_material"] = coll.surfs[iface][0]
						material["GTAMAT_coll_light"] = coll.surfs[iface][1]
						m.materials.append( material )
					face.material_index = cmat_dict[mat]
				m.update()
				o = bpy.data.objects.new( m.name, m )
				bpy.context.scene.objects.link( o )
				o.parent = coll_dummy
				coll_objs.append( o )
				
				## Create a Shadow Mesh
				if 0x334c4f43 == coll.fourcc and coll.flag_shadow:
					m = bpy.data.meshes.new( "shadow_mesh" )
					m.vertices.add( len( coll.sverts ) // 3 )
					m.vertices.foreach_set( "co", coll.sverts )
					m.faces.add( len( coll.sfaces ) // 4 )
					m.faces.foreach_set( "vertices_raw", coll.sfaces )
					material = bpy.data.materials.new( "Mat_shadow" )
					material.diffuse_intensity = 0.2
					m.materials.append( material )
					for iface, face in enumerate( m.faces ):
						mat = coll.ssurfs[iface]
						if not( mat in cmat_dict ):
							material = bpy.data.materials.new( "Mat_collision" )
							material.diffuse_intensity = 0.2
							cmat_dict[mat] =len( m.materials )
							rna_idprop_ui_prop_get( material, "GTA_Material" , create=True )
							rna_idprop_ui_prop_get( material, "GTAMAT_coll_light" , create=True )
							material["GTA_Material"]      = "collision"
							material["GTAMAT_coll_material"] = coll.ssurfs[iface][0]
							material["GTAMAT_coll_light"] = coll.ssurfs[iface][1]
							m.materials.append( material )
						face.material_index = cmat_dict[mat]
					m.update()
					o = bpy.data.objects.new( m.name, m )
					bpy.context.scene.objects.link( o )
					o.parent = coll_dummy
					coll_objs.append( o )
					rna_idprop_ui_prop_get( o, "GTA_Material" , create=True )
					o["GTA_Material"] = "shadow"
		
		### Grouping
		if None == map:
			g = bpy.data.groups.new( clump.name )
			if dff_props.char_skin:
				g.objects.link( skin_objs[skin_obj_index] )
				g.objects.link( arm_objs[skin_obj_index] )
			else:
				for obj in objs: g.objects.link( obj )
				if dff_props.vehicle:
					for obj in coll_objs: g.objects.link( obj )
			groups.append( g )
		
		### Set Layer
		## set a view layer of imported onjects to '0', for now,
		## cause i cant get the method to get "active view layer".
		#cur_layers = bpy.context.scene.layers[:]
		#bpy.context.scene.layers = [True]*20
		#if dff_props.char_skin:
		#	SetLayer(skin_objs[skin_obj_index], [0])
		#	SetLayer(arm_objs [skin_obj_index] , [0])
		#else:
		#	for obj in objs:
		#		if obj.type == 'MESH': SetLayer(obj, [0])
		#		else:                  SetLayer(obj, [0])
		#bpy.context.scene.update()
		#bpy.context.scene.layers = cur_layers
		
		## Set World Rot/Pos, Set Custom Properties
		for obj in objs:
			if None == obj.parent:
				rna_idprop_ui_prop_get( obj, "internal_pos", create=True )
				rna_idprop_ui_prop_get( obj, "internal_rot", create=True )
				obj["internal_pos"] = obj.location
				obj["internal_rot"] = obj.rotation_quaternion
				obj.location += dff.pos
				obj.rotation_quaternion.rotate( dff.rot.inverted() )
				#if None != txd:
				#	rna_idprop_ui_prop_get( obj, "txd", create=True )
				#	obj["txd"] = txd.img.name + "\\" + txd.name
		
		bpy.context.scene.update()
	return objs

def ImportDFF( inst = None, map = None ):
	dff_props = bpy.context.scene.gta_tools.dff_props
	global script_info
	if dff_props.dump_stat: print( "\n-----\n" + script_info + "\n-----" )
	
#	try:
	dff = ClassDFF()
	if dff_props.map:
		dff_props.char_skin = False
		dff_props.vehicle   = False
		dff_props.dump_stat = False
	else:
		if "CHAR" == dff_props.imp_type:
			dff_props.char_skin = True
			dff_props.vehicle   = False
		elif "VEHICLE" == dff_props.imp_type:
			dff_props.char_skin = False
			dff_props.vehicle   = True
		else:
			dff_props.char_skin = False
			dff_props.vehicle   = False
		dff_props.dump_stat = True
	
	if None == inst:
		filepath = dff_props.imp_filepath
		dff.name = os.path.splitext( os.path.split( filepath )[1] )[0]
		dff.ini  = 0
		dff.end  = os.path.getsize( filepath )
		txds      = None
	else:
		filepath = inst.ideobj.dff.img.path
		dff.name = inst.ideobj.dff.name
		dff.ini  = inst.ideobj.dff.ini
		dff.end  = inst.ideobj.dff.end
		dff.pos  = inst.pos
		dff.rot  = inst.rot
		dff.skip_nodes = map.skip_nodes
		txds     = [inst.ideobj.txd]
		if None != inst.nonlod: txds.append( inst.nonlod.ideobj.txd )
	
	try:
		file = open( filepath, "rb" )
		if dff_props.dump_stat:
			print( "Source DFF File : " + filepath )
			print( "File Size : " + str( dff.end ) + " bytes" )
	except:
		bpy.context.scene.gta_tools.msg +=  "Error : failed to open " + filepath + "\n"
		return
	
	file.seek( dff.ini, 0 )
	ReadDFF( file, dff )
	file.close()
	#return
	
	if None != txds: CheckTXDs( dff, txds, map )
	objs = CreateObject( dff, map )
	
	if dff_props.dump_stat: bpy.context.scene.gta_tools.msg +=  "Finished : Successfully.\n"
	
	return objs
		
#	except:
#		if dff_props.dump_stat: bpy.context.scene.gta_tools.msg +=  "Finished : With an ImportError.\n"


